/* Test __libc_res_nsend buffer mismanagement, additional TCP coverage.
   Copyright (C) 2016 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.  */

#include "tst-resolv-aux.h"

#include <errno.h>
#include <netdb.h>
#include <pthread.h>
#include <resolv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;

/* Server behavior tweaks.  */
static bool force_tcp_a;
static bool force_tcp_aaaa;
static int initial_address_count;
static int subsequent_address_count;
enum first_response_treatment
  {
    first_pass,
    first_zero_length,
    first_garbage,
    first_header_only,
    first_header_garbage,
    first_answer_garbage,
    first_servfail,
    first_drop,
    last_first_option = first_drop
  };
enum first_response_treatment first_a = first_servfail;
enum first_response_treatment first_aaaa = first_servfail;

/* Server state.  */
static bool response_tcp_forced_a;
static bool response_tcp_forced_aaaa;
static int response_number_a = 0;
static int response_number_aaaa = 0;

static void
reset_response_number (void)
{
  resolv_xlock (&lock);
  response_tcp_forced_a = false;
  response_tcp_forced_aaaa = false;
  response_number_a = 0;
  response_number_aaaa = 0;
  resolv_xunlock (&lock);
}

struct response_treatment
{
  /* The resonse code to use in the response packet header.  */
  int rcode;
  /* The number of addresses (in the additional section) for the name
     server record (in the authoritative section).  */
  int address_count;
  /* If true, return without sending any data.  This causes a
     zero-length response to be sent.  */
  bool no_data;
  /* If true, only send the 12-byte DNS header.  */
  bool header_only;
  /* If true, add garbage after a valid question section.  */
  bool garbage_answer;
  /* Spoofed ancount initial value.  */
  unsigned short ancount;
};

static void
add_garbage (struct response_builder *b)
{
  char buf[200];
  memset (buf, 0xc0, sizeof (buf));
  response_add_data (b, buf, sizeof (buf));
}

static struct response_treatment
response_treatment (struct response_builder *b,
                     int *response_number,
                     enum first_response_treatment treatment)
{
  ++*response_number;
  if (*response_number == 1)
    {
      switch (treatment)
        {
        case first_pass:
          return (struct response_treatment)
            {.rcode = 0, .address_count = initial_address_count};
          break;
        case first_garbage:
          add_garbage (b);
          return (struct response_treatment) {.no_data = true};
        case first_header_only:
          return (struct response_treatment) {.header_only = true};
          break;
        case first_header_garbage:
          response_init (b, (struct response_flags) {.qdcount = 1});
          add_garbage (b);
          return (struct response_treatment) {.no_data = true};
        case first_answer_garbage:
          return (struct response_treatment)
            {.garbage_answer = true, .ancount = 1};
        case first_zero_length:
          return (struct response_treatment) {.no_data = true};
        case first_servfail:
          return (struct response_treatment)
            {.rcode = 2, .address_count = initial_address_count};
        case first_drop:
          response_drop (b);
          response_close (b);
          return (struct response_treatment)
            {.rcode = 0, .address_count = 0};
        }
      abort ();
    }
  else
    return (struct response_treatment)
      {.rcode = 0, .address_count = subsequent_address_count};
}

static void
response (const struct response_context *ctx,
          struct response_builder *b,
          const char *qname, uint16_t qclass, uint16_t qtype)
{
  if (qname == NULL)
    {
      printf ("error: unparseable query\n");
      abort ();
    }

  /* If not using TCP, force its use if requested.  */
  bool force_tcp = false;
  resolv_xlock (&lock);
  if (!ctx->tcp)
    {
      if (qtype == T_A && force_tcp_a && !response_tcp_forced_a)
        {
          response_tcp_forced_a = true;
          force_tcp = true;
        }
      if (qtype == T_AAAA && force_tcp_aaaa && !response_tcp_forced_aaaa)
        {
          response_tcp_forced_aaaa = true;
          force_tcp = true;
        }
    }
  resolv_xunlock (&lock);
  if (force_tcp)
    {
      struct response_flags flags = {.tc = true};
      response_init (b, flags);
      response_add_question (b, qname, qclass, qtype);
      return;
    }

  resolv_xlock (&lock);
  struct response_treatment treatment;
  switch (qtype)
    {
    case T_A:
      treatment = response_treatment (b, &response_number_a, first_a);
      break;
    case T_AAAA:
      treatment = response_treatment (b, &response_number_aaaa, first_aaaa);
      break;
    default:
      printf ("error: unexpected QTYPE: %u\n", qtype);
      abort ();
    }
  resolv_xunlock (&lock);

  if (treatment.no_data)
    {
      /* Avoid running into an indefinite TCP hang in the
         first_garbage case.  */
      if (ctx->tcp && response_length (b) > 0)
        response_close (b);
      return;
    }

  struct response_flags flags =
    {
      .rcode = treatment.rcode,
      .ancount = treatment.ancount
    };
  response_init (b, flags);
  if (treatment.header_only)
    return;
  response_add_question (b, qname, qclass, qtype);
  if (treatment.garbage_answer)
    {
      add_garbage (b);
      return;
    }

  /* Only add the address record to the answer section if we requested
     any name server addresses.  */
  if (treatment.address_count > 0)
    {
      response_section (b, section_answer);

      response_open_record (b, qname, qclass, qtype, 0);
      switch (qtype)
        {
        case T_A:
          {
            char ipv4[4] = {10, response_number_a >> 8, response_number_a, 0};
            ipv4[3] = 2 * ctx->tcp + 4 * ctx->server_index;
            response_add_data (b, &ipv4, sizeof (ipv4));
          }
          break;
        case T_AAAA:
          {
            char ipv6[16]
              = {0x20, 0x01, 0xd, 0xb8, 0, 0, 0, 0, 0, 0, 0, 0,
                 response_number_aaaa >> 8, response_number_aaaa, 0, 0};
            ipv6[15] = 2 * ctx->tcp + 4 * ctx->server_index;
            response_add_data (b, &ipv6, sizeof (ipv6));
          }
          break;
        default:
          printf ("error: unexpected QTYPE: %s/%u/%u\n",
                  qname, qclass, qtype);
          abort ();
        }
      response_close_record (b);

      /* Add the name server record.  */
      response_section (b, section_authoritative);
      response_open_record (b, "example", C_IN, T_NS, 0);
      response_add_name (b, "ns.example");
      response_close_record (b);

      /* Increase the response size with name server addresses.  These
         addresses are not copied out of nss_dns, and thus do not
         trigger getaddrinfo retries with a larger buffer, making
         testing more predictable.  */
      response_section (b, section_additional);
      for (int i = 1; i <= treatment.address_count; ++i)
        {
          response_open_record (b, "ns.example", qclass, qtype, 0);
          switch (qtype)
            {
            case T_A:
              {
                char ipv4[4] = {10, i >> 8, i, 0};
                ipv4[3] = 2 * ctx->tcp + 4 * ctx->server_index;
                response_add_data (b, &ipv4, sizeof (ipv4));
              }
              break;
            case T_AAAA:
              {
                char ipv6[16]
                  = {0x20, 0x01, 0xd, 0xb8, 0, 0, 0, 0, 0, 0, 0, 0,
                     i >> 8, i, 0, 0};
                ipv6[15] = 2 * ctx->tcp + 4 * ctx->server_index;
                response_add_data (b, &ipv6, sizeof (ipv6));
              }
              break;
            default:
              printf ("error: unexpected QTYPE: %s/%u/%u\n",
                      qname, qclass, qtype);
              abort ();
            }
          response_close_record (b);
        }
    }
}

static char *
expected_result (unsigned port)
{
  struct resolv_memstream mem;
  resolv_xopen_memstream (&mem);
  int server_index;
  if (first_a == first_pass)
    /* No failure on first query.  Use first server. */
    server_index = 0;
  else
    server_index = 1;
  int response_number = server_index + 1;
  fprintf (mem.out, "address: STREAM/TCP 10.%u.%u.%u %u\n",
           (response_number >> 8) & 0xff, response_number & 0xff,
           force_tcp_a * 2 + server_index * 4, port);
  if (first_aaaa == first_pass)
    fprintf (mem.out, "address: STREAM/TCP 2001:db8::%x:%x %u\n",
             response_number & 0xffff,
             force_tcp_aaaa * 2 + server_index * 4, port);
  resolv_xfclose_memstream (&mem);
  return mem.buffer;
}

static void
test_one_size (const char *name)
{
  printf ("info: name: %s\n", name);
  printf ("info: initial_address_count: %d\n", initial_address_count);
  printf ("info: subsequent_address_count: %d\n", subsequent_address_count);
  reset_response_number ();

  struct addrinfo hints =
    {
      .ai_family = AF_UNSPEC,
      .ai_socktype = SOCK_STREAM,
      .ai_protocol = IPPROTO_TCP,
    };
  struct addrinfo *ai;
  int ret = getaddrinfo (name, "80", &hints, &ai);
  char *expected = expected_result (80);
  char *actual = format_addrinfo (ai, ret);
  if (ret == 0)
    freeaddrinfo (ai);
  free (expected);
  free (actual);
}

static void
test_different_sizes (void)
{
  /* Check that small sizes work.  */
  initial_address_count = 2;
  subsequent_address_count = 3;
  test_one_size ("www.example");

  /* This magic number produces a response size close to 2048
     bytes.  */
  initial_address_count = 124;
  subsequent_address_count = 2000;
  test_one_size ("www.example");

  /* Push the response size across 2048 bytes.  */
  test_one_size ("www1.example");
  test_one_size ("www12.example");
  test_one_size ("www123.example");
}

int
main (void)
{
  struct resolv_aux *aux = resolv_redirect
    ((struct redirect_config)
     {
       .response_callback = response
     });

  for (int f_t_a = 0; f_t_a <= 1; ++f_t_a)
    {
      force_tcp_a = f_t_a;
      for (int f_t_aaaa = 0; f_t_aaaa <= 1; ++f_t_aaaa)
        {
          force_tcp_aaaa = f_t_aaaa;
          for (int f_a = 0; f_a <= last_first_option; ++f_a)
            {
              first_a = f_a;
              for (int f_aaaa = 0; f_aaaa <= last_first_option; ++f_aaaa)
                {
                  first_aaaa = f_aaaa;

                  test_different_sizes ();

                  _res.options |= RES_SNGLKUP;
                  test_different_sizes ();

                  _res.options |= RES_SNGLKUPREOP;
                  test_different_sizes ();

                }
            }
          }
    }

  resolv_redirect_finish (aux);
}
