/* Test resolver behavior close to the maximum packet size.
   Copyright (C) 2016 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.  */

/* This test fills DNS responses in the authoriative section, so that
   it does not test buffer handling on the getaddrinfo/nss_dns or
   application/NSS boundary.  */

#include "tst-resolv-aux.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <resolv.h>
#include <netdb.h>

static pthread_mutex_t lock;
static struct config
{
  unsigned packet_size;
  bool force_tcp;
  unsigned response_number;
  unsigned char *response_sent;
} config;

static void
response (const struct response_context *ctx,
          struct response_builder *b,
          const char *qname, uint16_t qclass, uint16_t qtype)
{
  resolv_xlock (&lock);
  ++config.response_number;
  struct config cfg = config;
  resolv_xunlock (&lock);

  if (cfg.force_tcp && !ctx->tcp)
    {
      response_init (b, (struct response_flags) {.tc = true});
      response_add_question (b, qname, qclass, qtype);
      return;
    }

  response_init (b, (struct response_flags) {});
  response_add_question (b, qname, qclass, qtype);
  response_section (b, section_answer);
  response_open_record (b, qname, qclass, qtype, 0);
  switch (qtype)
    {
    case T_A:
      {
        char ipv4[4] = {10, 0, cfg.response_number, 0};
        ipv4[3] += 2 * ctx->tcp + 4 * ctx->server_index;
        response_add_data (b, &ipv4, sizeof (ipv4));
      }
      break;
    case T_AAAA:
      {
        char ipv6[16]
          = {0x20, 0x01, 0xd, 0xb8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
        ipv6[13] = cfg.response_number;
        ipv6[15] += 2 * ctx->tcp + 4 * ctx->server_index;
        response_add_data (b, &ipv6, sizeof (ipv6));
      }
      break;
    default:
      printf ("error: unexpected QTYPE: %s/%u/%u\n",
              qname, qclass, qtype);
      abort ();
    }
  response_close_record (b);

  /* Fill the authoritative section with NS records, exactly up to the
     requested packet size.  */
  response_section (b, section_authoritative);
  int count = 0;                /* counter for unique domain names */
  while (true)
    {
      const size_t record_header_size
        = 1                     /* label length */
        + 1                     /* label terminator */
        + 2                     /* RTYPE */
        + 2                     /* RCLASS */
        + 4                     /* TTL */
        + 2                     /* RDATA length */
        + 1                     /* label length (in RDATA) */
        + 1;                    /* label terminator */
      size_t current = response_length (b);
      if (current > cfg.packet_size)
        /* Something went wrong with the size adjustment.  */
        abort ();
      unsigned remaining = cfg.packet_size - current;
      if (remaining == 0)
        break;
      else if (remaining < record_header_size)
        /* This cannot happen, as it would invalidate the test.  */
        abort ();
      size_t to_distribute;
      if (remaining <= record_header_size + 2 * 63)
        to_distribute = remaining - record_header_size;
      else
        /* Avoid a remaining tail which is larger than
           record_header_size.  */
        to_distribute = 63;
      int first_label_length = to_distribute / 2;
      int second_label_length = to_distribute - first_label_length;
      char buf[64];
      ++count;
      snprintf (buf, sizeof (buf), "%0*d", first_label_length, count);
      response_open_record (b, buf, C_IN, T_NS, 86400);
      ++count;
      snprintf (buf, sizeof (buf), "%0*d", second_label_length, count);
      response_add_name (b, buf);
      response_close_record (b);
    }

  unsigned char *response_sent = response_buffer (b);
  resolv_xlock (&lock);
  free (config.response_sent);
  config.response_sent = response_sent;
  resolv_xunlock (&lock);
}

enum
  {
    udp_max_packet = 65535 - 20 /* IPv4 header */ - 8 /* UDP header */
  };

static void
test_resolution (size_t requested, bool tcp)
{
  resolv_xlock (&lock);
  free (config.response_sent);
  config = (struct config)
    {
      .packet_size = requested,
      .force_tcp = tcp,
      .response_number = 0,
    };
  resolv_xunlock (&lock);

  const char *expected;
  if (config.force_tcp)
    expected = "name: www.example\n"
      "address: 10.0.2.2\n";
  else
    expected = "name: www.example\n"
      "address: 10.0.1.0\n";

  struct hostent *e = gethostbyname ("www.example");
  check_hostent (e, expected);

  if (config.force_tcp)
    expected = "name: www.example\n"
      "address: 2001:db8::2:2\n";
  else
    expected = "name: www.example\n"
      "address: 2001:db8::1:0\n";
  resolv_xlock (&lock);
  config.response_number = 0;
  resolv_xunlock (&lock);
  e = gethostbyname2 ("www.example", AF_INET6);
  check_hostent (e, expected);

  struct addrinfo hints =
    {
      .ai_family = AF_INET,
      .ai_protocol = IPPROTO_TCP,
      .ai_socktype = SOCK_STREAM
    };
  if (config.force_tcp)
    expected = "address: STREAM/TCP 10.0.2.2 80\n";
  else
    expected = "address: STREAM/TCP 10.0.1.0 80\n";
  resolv_xlock (&lock);
  config.response_number = 0;
  resolv_xunlock (&lock);
  struct addrinfo *ai;
  int ret = getaddrinfo ("www.example", "80", &hints, &ai);
  check_addrinfo (ai, ret, expected);
  if (ret == 0)
    freeaddrinfo (ai);

  hints.ai_family = AF_INET6;
  if (config.force_tcp)
    expected = "address: STREAM/TCP 2001:db8::2:2 80\n";
  else
    expected = "address: STREAM/TCP 2001:db8::1:0 80\n";
  resolv_xlock (&lock);
  config.response_number = 0;
  resolv_xunlock (&lock);
  ret = getaddrinfo ("www.example", "80", &hints, &ai);
  check_addrinfo (ai, ret, expected);
  if (ret == 0)
    freeaddrinfo (ai);
}

static void
test_size (size_t requested, int skew, int rtype, bool tcp)
{
  resolv_xlock (&lock);
  free (config.response_sent);
  config = (struct config)
    {
      .packet_size = requested,
      .force_tcp = tcp,
      .response_number = 0,
    };
  resolv_xunlock (&lock);

  unsigned char *buffer = malloc (requested + skew);
  if (buffer == NULL)
    {
      printf ("error: malloc: %m\n");
      abort ();
    }
  int ret = res_query ("ww1.example", C_IN, rtype, buffer, requested);
  if (ret <= 0)
    {
      printf ("error: res_send for size %zu: %d (%m)\n", requested, ret);
      abort ();
    }
  if ((size_t) ret != requested)
    {
      printf ("error: res_send requested %zu bytes, but got %d\n",
              requested, ret);
      abort ();
    }

  /* Check that the response buffer passed through libresolv
     unchanged.  */
  unsigned char *response_sent;
  resolv_xlock (&lock);
  response_sent = config.response_sent;
  config.response_sent = NULL;
  resolv_xunlock (&lock);
  if (memcmp (response_sent, buffer, requested) != 0)
    {
      printf ("error: response buffer modified in flight\n");
      abort ();
    }

  free (response_sent);
  free (buffer);

  /* Check with buffer that is too short.  */
  buffer = malloc (requested - 1 + skew);
  if (buffer == NULL)
    {
      printf ("error: malloc: %m\n");
      abort ();
    }
  ret = res_query ("ww1.example", C_IN, rtype, buffer, requested - 1);
  if (ret <= 0)
    {
      printf ("error: res_send for size %zu: %d (%m)\n", requested - 1, ret);
      abort ();
    }
  int expected_ret;
  if (tcp)
    /* In the TCP case, res_query will return the server-supplied
       length, even if the buffer size is too small..  */
    expected_ret = requested;
  else
    /* In the UDP case, the kernel has simply discarded the remaining
       packet, and we do not know that truncation has occured.  */
    expected_ret = requested - 1;
  if (ret != expected_ret)
    {
      printf ("error: res_send requested %d bytes, but got %d (truncation)\n",
              expected_ret, ret);
      abort ();
    }
  resolv_xlock (&lock);
  response_sent = config.response_sent;
  config.response_sent = NULL;
  resolv_xunlock (&lock);

  if (tcp)
    {
      /* libresolv should have detected truncation.  */
      if ((buffer[2] & 0x02) == 0)
        {
          printf ("error: libresolv not set TC bit"
                  " on client-side truncation\n");
        }
      /* Clear TC bit, to match original server response.  */
      buffer[2] &= ~0x02;
      if (memcmp (response_sent, buffer, requested - 1) != 0)
        {
          printf ("error: response buffer modified in flight"
                  " (client truncation)\n");
          abort ();
        }
    }
  else
    /* UDP.  libresolv could not detect truncation.  */
    if (memcmp (response_sent, buffer, requested - 1) != 0)
      {
        printf ("error: response buffer modified in flight (truncation)\n");
        abort ();
      }

  free (response_sent);
  free (buffer);
}

static void
test_sizes (void)
{
  for (int tcp = 0; tcp < 2; ++tcp)
    {
      for (int skew = 0; skew < 2; ++skew)
        for (int rtype = T_A; rtype <= T_AAAA; rtype += T_AAAA - T_A)
          {
            test_size (512, skew, rtype, tcp);
            test_size (2048, skew, rtype, tcp);
            test_size (udp_max_packet, skew, rtype, tcp);
          }

      test_resolution (512, tcp);
      test_resolution (2048, tcp);
      test_resolution (udp_max_packet, tcp);
    }
  for (int skew = 0; skew < 2; ++skew)
    {
      test_size (65535, skew, T_A, true);
      test_size (65535, skew, T_AAAA, true);
    }
  test_resolution (65535, true);
}

int
main (void)
{
  struct resolv_aux *aux = resolv_redirect
    ((struct redirect_config)
     {
       .response_callback = response
     });

  /* With valgrind, large packet construction can take some time.  We
     do not drop any packets, so the timeout value does not affect run
     time, so we simply increase the timeout.  */
  _res.retrans = 30;
  test_sizes ();

  resolv_redirect_finish (aux);
}
