/* Test case for combined A/AAAA responses with partial failures.
   Copyright (C) 2016 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.  */

/* The goal of this test is to check how DNS errors (timeouts, garbled
   packets, DNS errors) are handled and which responses/servers are
   selected.  */

#include "tst-resolv-aux.h"

#include <errno.h>
#include <netdb.h>
#include <pthread.h>
#include <resolv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>

static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;

/* Server behavior control.  */
enum
  {
    rcode_nodata = -4,         /* RCODE=0, but no address */
    rcode_drop = -3,            /* drop packet */
    rcode_zero_length = -2,     /* empty packet */
    rcode_garbage = -1,         /* packet with garbage */
    /* RCODEs 0 to 15 according to the DNS specification.  */
  };
static int first_rcode[] =
  {
    [T_A] = 0,
    [T_AAAA] = 0,
  };
static int response_count[] =
  {
    [T_A] = 0,
    [T_AAAA] = 0,
  };

static void
init_round (int first_a, int first_aaaa)
{
  resolv_xlock (&lock);
  first_rcode[T_A] = first_a;
  first_rcode[T_AAAA] = first_aaaa;
  memset (response_count, 0, sizeof (response_count));
  resolv_xunlock (&lock);
}


static void
response (const struct response_context *ctx,
          struct response_builder *b,
          const char *qname, uint16_t qclass, uint16_t qtype)
{
  if (qname == NULL)
    {
      printf ("error: unparseable query\n");
      abort ();
    }
  if (qtype != T_A && qtype != T_AAAA)
    {
      printf ("error: invalid QTYPE: %d\n", qtype);
      abort ();
    }

  resolv_xlock (&lock);
  int rcount = ++response_count[qtype];
  bool first_response = rcount == 1;
  int rcode;
  if (first_response)
    rcode = first_rcode[qtype];
  else
    rcode = 0;
  resolv_xunlock (&lock);

  if (first_response)
    {
      /* Handle special-case RCODEs.  */
      switch (rcode)
        {
        case rcode_nodata:
          response_init (b, (struct response_flags) {});
          response_add_question (b, qname, qclass, qtype);
          return;
        case rcode_drop:
          response_drop (b);
          return;
        case rcode_zero_length:
          return;
        case rcode_garbage:
          {
            char data[12];
            memset (data, 0xff, sizeof (data));
            response_add_data (b, data, sizeof (data));
            return;
          }
        }
      /* Default: Fall through, send as regular DNS rcode.  */
    }

  struct response_flags flags = {.rcode = rcode};
  response_init (b, flags);
  response_add_question (b, qname, qclass, qtype);

  /* Do not include data in error responses, better matching actual
     server behavior (although nss_dns currently does not extract the
     negative TTL from the packet, which is a bug).  Another test
     could make sure that libresolv actually ignores data in RCODE !=
     0 responses (it does not).  But recursive resolvers are not
     expected to send such responses.

     We might want to include SOA responses here eventually.  */
  if (rcode > 0)
    return;

  response_section (b, section_answer);
  response_open_record (b, qname, qclass, qtype, 0);
  switch (qtype)
    {
    case T_A:
      {
        char ipv4[4] =
          {
            rcount, rcode, 0,
            2 * ctx->tcp + 4 * ctx->server_index
          };
        response_add_data (b, &ipv4, sizeof (ipv4));
      }
      break;
    case T_AAAA:
      {
        char ipv6[16] =
          {
            0x20, 0x01, 0xd, 0xb8, 0, 0, 0, 0, 0, 0, 0, 0,
            rcount >> 8, rcount,
            0, 2 * ctx->tcp + 4 * ctx->server_index
          };
        response_add_data (b, &ipv6, sizeof (ipv6));
      }
      break;
    }
  response_close_record (b);
}

static char *
one_test (int first_a, int first_aaaa)
{
  struct resolv_aux *aux = resolv_redirect
    ((struct redirect_config)
     {
       .response_callback = response,
       .single_thread_udp = true
     });

  init_round (first_a, first_aaaa);
  printf ("info: test case %d/%d\n", first_a, first_aaaa);
  struct addrinfo hints =
    {
      .ai_family = AF_UNSPEC,
      .ai_socktype = SOCK_STREAM,
      .ai_protocol = IPPROTO_TCP,
    };
  struct addrinfo *ai;
  int ret = getaddrinfo ("www.example", "80", &hints, &ai);
  resolv_redirect_finish (aux);

  char *result = format_addrinfo (ai, ret);
  for (char *p = result; *p != '\0'; ++p)
    if (*p == '\n')
      *p = ' ';
  if (ret == 0)
    freeaddrinfo (ai);
  return result;
}

/* If true, write a new output log file to the expected-new file.  */
static bool write_log = false;

int
main (void)
{
  FILE *out;
  if (write_log)
    {
      out = fopen ("tst-partial-fail.expected-new", "w");
      if (out == NULL)
        {
          printf ("error: could not open \"log\": %m\n");
          return 1;
        }

      for (int first_a = rcode_nodata; first_a <= 15; ++first_a)
        for (int first_aaaa = rcode_nodata; first_aaaa <= 15; ++first_aaaa)
          {
            char *result = one_test (first_a, first_aaaa);
            fprintf (out, "%d %d %s\n", first_a, first_aaaa, result);
            free (result);
          }
      if (ferror (out) || fclose (out) < 0)
        {
          printf ("error: log wite failure\n");
          abort ();
        }
      return 0;
    }

  FILE *in = fopen ("tst-partial-fail.expected", "r");
  if (in == NULL)
    {
      printf ("error: could not open tst-partial-fail.expected: %m\n");
      return 1;
    }
  char *line_buffer = NULL;
  size_t line_length = 0;

  int errors = 0;
  while (!feof (in))
    {
      if (getline (&line_buffer, &line_length, in) <= 0)
        break;
      {
        size_t len = strlen (line_buffer);
        if (len == 0 || line_buffer[len - 1] != '\n')
          {
            printf ("error: missing newline in input\n");
            abort ();
          }
        line_buffer[len - 1] = '\0';
      }

      int first_a;
      int first_aaaa;
      char space;
      int pos;
      if (sscanf (line_buffer, "%d %d%c%n",
                  &first_a, &first_aaaa, &space, &pos) != 3
          || space != ' ')
        {
          printf ("error: could not parse log line: %s\n", line_buffer);
          ++errors;
        }
      char *result = one_test (first_a, first_aaaa);
      if (strcmp (result, line_buffer + pos) != 0)
        {
          printf ("- %d %d %s\n", first_a, first_aaaa, line_buffer + pos);
          printf ("+ %d %d %s\n", first_a, first_aaaa, result);
          ++errors;
        }
      free (result);
    }
  if (ferror (in) || fclose (in) < 0)
    {
      printf ("error: reading log file\n");
      abort ();
    }
  free (line_buffer);

  if (errors > 0)
    {
      printf ("error: %d errors encountered\n", errors);
      abort ();
    }
};
