/* Support functionality for stub resolver testing.
   Copyright (C) 2016 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.  */

#include "tst-resolv-aux.h"

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <nss.h>
#include <pthread.h>
#include <resolv.h>
#include <search.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <unistd.h>

/* Print additional debugging data.  */
static bool verbose = true;

/* Print DNS transaction IDs in debugging data.  Introduces spurious
   diffs.  */
static bool verbose_with_txnid = false;

/* Counter with the number of reported errors (which did not lead to
   an abort).  */
static int errors;
static pthread_mutex_t errors_lock = PTHREAD_MUTEX_INITIALIZER;

/* Response builder. */

enum
  {
    max_response_length = 65536
  };

struct to_be_freed
{
  struct to_be_freed *next;
  void *ptr;
};

struct response_builder
{
  const unsigned char *query_buffer;
  size_t query_length;

  unsigned char buffer[max_response_length];
  size_t offset;
  enum response_section section;
  bool drop;                    /* Discard generated response.  */
  bool close;                   /* Close TCP client connection.  */

  /* Offset of the two-byte RDATA length field in the currently
     written RDATA sub-structure.  0 if no RDATA is being written.  */
  size_t current_rdata_offset;

  /* Hash table for locating targets for label compression.  */
  struct hsearch_data compression_offsets;
  /* List of pointers which need to be freed.  Used for domain names
     involved in label compression.  */
  struct to_be_freed *to_be_freed;
};

static struct response_builder *response_builder_allocate
  (const unsigned char *query_buffer, size_t query_length);
static void response_builder_free (struct response_builder *);

/* General-purpose helper routines. */

static void record_error (struct resolv_aux *);

static void *
xmalloc (size_t size)
{
  void *result = malloc (size);
  if (result == NULL)
    {
      printf ("error: malloc (%zu): %m\n", size);
      abort ();
    }
  return result;
}

static int
xsocket (int type)
{
  int sock = socket (AF_INET, type, 0);
  if (sock < 0)
    {
      printf ("error: socket (type %d): %m\n", type);
      abort ();
    }
  return sock;
}

static void
xsendto (int sock, const void *buffer, size_t length,
         int flags, struct sockaddr *addr, socklen_t addrlen)
{
  ssize_t ret = sendto (sock, buffer, length, flags | MSG_NOSIGNAL,
                        addr, addrlen);
  if (ret < 0)
    {
      /* ENOBUFS is an expected error if we respond to a request and
         the socket is being closed.  */
      if (errno == ENOBUFS)
        printf ("warning: UDP send: %m\n");
      else
        {
          printf ("error: UDP send: %m\n");
          abort ();
        }
    }
  else if ((size_t) ret != length)
    {
      printf ("error: UDP short send: sent %zd bytes, but requested %zu\n",
              ret, length);
      abort ();
    }
}

static bool
read_fully (struct resolv_aux *aux, int fd, void *buf, size_t len, bool eof_ok)
{
  const void *const end = buf + len;
  while (buf < end)
    {
      ssize_t ret = read (fd, buf, end - buf);
      if (ret == 0)
        {
          if (!eof_ok)
            {
              printf ("error: unexpected EOF on TCP connection\n");
              record_error (aux);
            }
          return false;
        }
      else if (ret < 0)
        {
          if (!eof_ok || errno != ECONNRESET)
            {
              printf ("error: TCP read: %m\n");
              record_error (aux);
            }
          return false;
        }
      buf += ret;
      eof_ok = false;
    }
  return true;
}

static void
xwritev_fully (int fd, struct iovec *buffers, size_t count)
{
  while (count > 0)
    {
      /* Skip zero-length write requests.  */
      if (buffers->iov_len == 0)
        {
          ++buffers;
          --count;
          continue;
        }
      /* Try to rewrite the remaing buffers.  */
      ssize_t ret = writev (fd, buffers, count);
      if (ret < 0)
        {
          printf ("error: writev: %m\n");
          abort ();
        }
      if (ret == 0)
        {
          printf ("error: writev: invalid return value zero\n");
          abort ();
        }
      /* Find the buffers that were successfully written.  */
      while (ret > 0)
        {
          if (count == 0)
            {
              printf ("error: internal writev consistency failure\n");
              abort ();
            }
          /* Current buffer was partially written.  */
          if (buffers->iov_len > (size_t) ret)
            {
              buffers->iov_base += ret;
              buffers->iov_len -= ret;
              ret = 0;
            }
          else
            {
              ret -= buffers->iov_len;
              buffers->iov_len = 0;
              ++buffers;
              --count;
            }
        }
    }
}

static void
xjoin (pthread_t thread)
{
  int ret = pthread_join (thread, NULL);
  if (ret != 0)
    {
      errno = ret;
      printf ("error: pthread_join: %m\n");
      abort ();
    }
}

void
resolv_xlock (pthread_mutex_t *mutex)
{
  int ret = pthread_mutex_lock (mutex);
  if (ret != 0)
    {
      errno = ret;
      printf ("error: pthread_mutex_lock: %m\n");
      abort ();
    }
}


void
resolv_xunlock (pthread_mutex_t *mutex)
{
  int ret = pthread_mutex_unlock (mutex);
  if (ret != 0)
    {
      errno = ret;
      printf ("error: pthread_mutex_unlock: %m\n");
      abort ();
    }
}

/* DNS server implementation. */

struct query_info
{
  char qname[MAXDNAME];
  uint16_t qclass;
  uint16_t qtype;
};

static void
parse_query (struct query_info *info,
             const unsigned char *buffer, size_t length)
{
  if (length < 12)
    {
      printf ("error: malformed query: too short: %zu bytes",
              length);
      abort ();
    }

  int ret = dn_expand (buffer, buffer + length, buffer + 12,
                       info->qname, sizeof (info->qname));
  if (ret < 0)
    {
      printf ("error: malformed query: cannot uncompress QNAME\n");
      abort ();
    }

  /* Obtain QTYPE and QCLASS.  */
  size_t remaining = length - (12 + ret);
  struct
  {
    uint16_t qtype;
    uint16_t qclass;
  } qtype_qclass;
  if (remaining < sizeof (qtype_qclass))
    {
      printf ("error: malformed query: "
              "query lacks QCLASS/QTYPE, QNAME: %s\n", info->qname);
      abort ();
    }
  memcpy (&qtype_qclass, buffer + 12 + ret, sizeof (qtype_qclass));
  info->qclass = ntohs (qtype_qclass.qclass);
  info->qtype = ntohs (qtype_qclass.qtype);
}

struct resolv_aux_server
{
  struct sockaddr_in address;
  int socket_udp;
  int socket_tcp;
  size_t response_number;

  pthread_t thread_udp;
  pthread_t thread_tcp;
};

struct resolv_aux
{
  pthread_mutex_t lock;
  struct resolv_aux_server servers[max_servers];

  /* Used if config.single_thread_udp is true.  */
  pthread_t thread_udp_single;

  struct redirect_config config;
  int errors;
  bool termination_requested;
};

static void
record_error (struct resolv_aux *aux)
{
  resolv_xlock (&aux->lock);
  ++aux->errors;
  resolv_xunlock (&aux->lock);
}

/* Create UDP and TCP server sockets.  */
static void
make_server_sockets (struct resolv_aux_server *server)
{
  while (true)
    {
      server->socket_udp = xsocket (SOCK_DGRAM);
      server->socket_tcp = xsocket (SOCK_STREAM);

      /* Pick the address for the UDP socket.  */
      server->address = (struct sockaddr_in)
        {
          .sin_family = AF_INET,
          .sin_addr = {.s_addr = htonl (INADDR_LOOPBACK)}
        };
      if (bind (server->socket_udp,
                (struct sockaddr *)&server->address,
                sizeof (server->address)) != 0)
        {
          printf ("error: UDP bind: %m\n");
          abort ();
        }

      /* Retrieve the address. */
      socklen_t addrlen = sizeof (server->address);
      if (getsockname (server->socket_udp,
                       (struct sockaddr *)&server->address, &addrlen) != 0)
        {
          printf ("error: UDP getsockname: %m\n");
          abort ();
        }

      /* Bind the TCP socket to the same address.  */
      {
        int on = 1;
        if (setsockopt (server->socket_tcp, SOL_SOCKET, SO_REUSEADDR,
                        &on, sizeof (on)) != 0)
          {
            printf ("error: setsockopt (SO_REUSEADDR): %m\n");
            abort ();
          }
      }
      if (bind (server->socket_tcp,
                (struct sockaddr *)&server->address,
                sizeof (server->address)) != 0)
        {
          /* Port collision.  The UDP bind succeeded, but the TCP BIND
             failed.  We assume here that the kernel will pick the
             next local UDP address randomly.  */
          if (errno == EADDRINUSE)
            {
              close (server->socket_udp);
              close (server->socket_tcp);
              continue;
            }
          printf ("error: TCP bind: %m\n");
          abort ();
        }
      if (listen (server->socket_tcp, 5) != 0)
        {
          printf ("error: TCP listen: %m\n");
          abort ();
        }

      break;
    }
}

typedef void (*thread_callback) (struct resolv_aux *, int server_index);

struct thread_closure
{
  struct resolv_aux *aux;
  thread_callback callback;
  int server_index;
};

static void *
thread_callback_wrapper (void *arg)
{
  struct thread_closure *closure = arg;
  closure->callback (closure->aux, closure->server_index);
  free (closure);
  return NULL;
}

static void
start_server_thread (struct resolv_aux *aux, int server_index,
                     pthread_t *thr, thread_callback callback)
{
  struct thread_closure *closure = xmalloc (sizeof (*closure));
  *closure = (struct thread_closure)
    {
      .aux = aux,
      .callback = callback,
      .server_index = server_index,
    };
  int ret = pthread_create (thr, NULL, thread_callback_wrapper, closure);
  if (ret != 0)
    {
      errno = ret;
      printf ("error: pthread_create: %m\n");
      abort ();
    }
}

/* Process one UDP query.  Return false if a termination requested has
   been detected.  */
static bool
server_thread_udp_process_one (struct resolv_aux *aux, int server_index)
{
  unsigned char query[512];
  struct sockaddr_storage peer;
  socklen_t peerlen = sizeof (peer);
  ssize_t ret = recvfrom (aux->servers[server_index].socket_udp,
                          query, sizeof (query), 0,
                          (struct sockaddr *) &peer, &peerlen);
  if (ret < 0)
    {
      printf ("error: UDP recvfrom: %m\n");
      abort ();
    }

  /* Check for termination.  */
  {
    bool termination_requested;
    resolv_xlock (&aux->lock);
    termination_requested = aux->termination_requested;
    resolv_xunlock (&aux->lock);
    if (termination_requested)
      return false;
  }


  struct query_info qinfo;
  parse_query (&qinfo, query, ret);
  if (verbose)
    {
      if (verbose_with_txnid && ret >= 2)
        printf ("info: UDP server %d: incoming query:"
                " %zd bytes, %s/%u/%u, tnxid=0x%02x%02x\n",
                server_index, ret, qinfo.qname, qinfo.qclass, qinfo.qtype,
                query[0], query[1]);
      else
        printf ("info: UDP server %d: incoming query:"
                " %zd bytes, %s/%u/%u\n",
                server_index, ret, qinfo.qname, qinfo.qclass, qinfo.qtype);
    }

  struct response_context ctx =
    {
      .query_buffer = query,
      .query_length = ret,
      .server_index = server_index,
      .tcp = false,
    };
  struct response_builder *b = response_builder_allocate (query, ret);
  aux->config.response_callback
    (&ctx, b, qinfo.qname, qinfo.qclass, qinfo.qtype);

  if (b->drop)
    {
      if (verbose)
        printf ("info: UDP server %d: dropping response to %s/%u/%u\n",
                server_index, qinfo.qname, qinfo.qclass, qinfo.qtype);
    }
  else
    {
      if (verbose)
        {
          if (b->offset >= 12)
            printf ("info: UDP server %d: sending response:"
                    " %zu bytes, RCODE %d (for %s/%u/%u)\n",
                    server_index, b->offset, b->buffer[3] & 0x0f,
                    qinfo.qname, qinfo.qclass, qinfo.qtype);
          else
            printf ("info: UDP server %d: sending response: %zu bytes"
                    " (for %s/%u/%u)\n",
                    server_index, b->offset,
                    qinfo.qname, qinfo.qclass, qinfo.qtype);
        }
      xsendto (aux->servers[server_index].socket_udp,
               b->buffer, b->offset, 0,
               (struct sockaddr *) &peer, peerlen);
    }
  response_builder_free (b);
  return true;
}

static void
server_thread_udp (struct resolv_aux *aux, int server_index)
{
  while (server_thread_udp_process_one (aux, server_index))
    ;
}

static void *
server_thread_udp_single (void *closure)
{
  struct resolv_aux *aux = closure;

  struct pollfd fds[max_servers];
  for (int server_index = 0; server_index < max_servers; ++server_index)
    if (aux->config.servers[server_index].disable_udp)
      fds[server_index] = (struct pollfd) {.fd = -1};
    else
      {
        fds[server_index] = (struct pollfd)
          {
            .fd = aux->servers[server_index].socket_udp,
            .events = POLLIN
          };

        /* Make the socket non-blocking.  */
        int flags = fcntl (aux->servers[server_index].socket_udp, F_GETFL, 0);
        if (flags < 0)
          {
            printf ("error: fcntl (F_GETFL): %m\n");
            abort ();
          }
        flags |= O_NONBLOCK;
        if (fcntl (aux->servers[server_index].socket_udp, F_SETFL, flags) < 0)
          {
            printf ("error: fcntl (F_SETFL): %m\n");
            abort ();
          }
      }

  while (true)
    {
      int ret = poll (fds, max_servers, -1);
      if (ret < 0)
        {
          printf ("error: UDP poll: %m\n");
          abort ();
        }

      for (int server_index = 0; server_index < max_servers; ++server_index)
        if (fds[server_index].revents != 0)
          {
            if (!server_thread_udp_process_one (aux, server_index))
              goto out;
            fds[server_index].revents = 0;
          }
    }

 out:
  return NULL;
}

static void
start_server_thread_udp_single (struct resolv_aux *aux)
{
  int ret = pthread_create (&aux->thread_udp_single, NULL,
                            server_thread_udp_single, aux);
  if (ret != 0)
    {
      errno = ret;
      printf ("error: pthread_create: %m\n");
      abort ();
    }
}

struct tcp_thread_closure
{
  struct resolv_aux *aux;
  int server_index;
  int client_socket;
};

static void *
server_thread_tcp_client (void *arg)
{
  struct tcp_thread_closure *closure = arg;

  while (true)
    {
      /* Read packet length.  */
      uint16_t query_length;
      if (!read_fully (closure->aux, closure->client_socket,
                       &query_length, sizeof (query_length), true))
        break;
      query_length = ntohs (query_length);

      /* Read the packet.  */
      unsigned char *query_buffer = xmalloc (query_length);
      read_fully (closure->aux, closure->client_socket,
                  query_buffer, query_length, false);

      struct query_info qinfo;
      parse_query (&qinfo, query_buffer, query_length);
      if (verbose)
        {
          if (verbose_with_txnid && query_length >= 2)
            printf ("info: UDP server %d: incoming query:"
                    " %d bytes, %s/%u/%u, tnxid=0x%02x%02x\n",
                    closure->server_index, query_length,
                    qinfo.qname, qinfo.qclass, qinfo.qtype,
                    query_buffer[0], query_buffer[1]);
          else
            printf ("info: TCP server %d: incoming query:"
                    " %u bytes, %s/%u/%u\n",
                    closure->server_index, query_length,
                    qinfo.qname, qinfo.qclass, qinfo.qtype);
        }

      struct response_context ctx =
        {
          .query_buffer = query_buffer,
          .query_length = query_length,
          .server_index = closure->server_index,
          .tcp = true,
        };
      struct response_builder *b = response_builder_allocate
        (query_buffer, query_length);
      closure->aux->config.response_callback
        (&ctx, b, qinfo.qname, qinfo.qclass, qinfo.qtype);

      if (b->drop)
        {
          if (verbose)
            printf ("info: TCP server %d: dropping response to %s/%u/%u\n",
                    closure->server_index,
                    qinfo.qname, qinfo.qclass, qinfo.qtype);
        }
      else
        {
          if (verbose)
            printf ("info: TCP server %d: sending response: %zu bytes"
                    " (for %s/%u/%u)\n",
                    closure->server_index, b->offset,
                    qinfo.qname, qinfo.qclass, qinfo.qtype);
          uint16_t length = htons (b->offset);
          struct iovec buffers[2] =
            {
              {&length, sizeof (length)},
              {b->buffer, b->offset}
            };
          xwritev_fully (closure->client_socket, buffers, 2);
        }
      bool close_flag = b->close;
      response_builder_free (b);
      free (query_buffer);
      if (close_flag)
        break;
    }

  close (closure->client_socket);
  free (closure);
  return NULL;
}

static void
server_thread_tcp (struct resolv_aux *aux, int server_index)
{
  while (true)
    {
      /* Get the client conenction.  */
      int client_socket = accept
        (aux->servers[server_index].socket_tcp, NULL, NULL);
      if (client_socket < 0)
        {
          printf ("error: TCP accept: %m\n");
          abort ();
        }

      /* Check for termination.  */
      resolv_xlock (&aux->lock);
      if (aux->termination_requested)
        {
          resolv_xunlock (&aux->lock);
          close (client_socket);
          break;
        }
      resolv_xunlock (&aux->lock);

      /* Spawn a new thread for handling this connection.  */
      struct tcp_thread_closure *closure = malloc (sizeof (*closure));
      *closure = (struct tcp_thread_closure)
        {
          .aux = aux,
          .server_index = server_index,
          .client_socket = client_socket,
        };

      pthread_t thr;
      int ret = pthread_create (&thr, NULL, server_thread_tcp_client, closure);
      if (ret != 0)
        {
          errno = ret;
          printf ("error: pthread_create: %m");
          abort ();
        }
      ret = pthread_detach (thr);
      if (ret != 0)
        {
          errno = ret;
          printf ("error: pthread_detach: %m");
          abort ();
        }
    }
}

static void
resolv_redirect_once (void)
{
  /* Only use nss_dns.  */
  __nss_configure_lookup ("hosts", "dns");
}
pthread_once_t resolv_redirect_once_var = PTHREAD_ONCE_INIT;

static void
set_search_path (struct redirect_config config)
{
  memset (_res.defdname, 0, sizeof (_res.defdname));
  memset (_res.dnsrch, 0, sizeof (_res.dnsrch));

  char *current = _res.defdname;
  char *end = current + sizeof (_res.defdname);

  for (unsigned i = 0;
       i < sizeof (config.search) / sizeof (config.search[0]); ++i)
    {
      if (config.search[i] == NULL)
        continue;

      size_t length = strlen (config.search[i]) + 1;
      size_t remaining = end - current;
      if (length > remaining)
        {
          printf ("error: configured search path too long\n");
          abort ();
        }
      memcpy (current, config.search[i], length);
      _res.dnsrch[i] = current;
      current += length;
    }
}

struct resolv_aux *
resolv_redirect (struct redirect_config config)
{
  /* Apply configuration defaults.  */
  if (config.nscount == 0)
    config.nscount = max_servers;

  struct resolv_aux *aux = xmalloc (sizeof (*aux));
  *aux = (struct resolv_aux) {
    .config = config,
    .lock = PTHREAD_MUTEX_INITIALIZER,
  };

  /* Perform one-time initialization of NSS.  */
  {
    int ret = pthread_once (&resolv_redirect_once_var, resolv_redirect_once);
    if (ret != 0)
      {
        errno = ret;
        printf ("error: pthread_once: %m\n");
        abort ();
      }
  }

  /* Create all the servers, to reserve the necessary ports.  */
  for (int server_index = 0; server_index < config.nscount; ++server_index)
    make_server_sockets (aux->servers + server_index);

  /* Start server threads.  Disable the server ports, as
     requested.  */
  for (int server_index = 0; server_index < config.nscount; ++server_index)
    {
      struct resolv_aux_server *server = aux->servers + server_index;
      if (config.servers[server_index].disable_udp)
        {
          close (server->socket_udp);
          server->socket_udp = -1;
        }
      else if (!config.single_thread_udp)
        start_server_thread (aux, server_index,
                             &server->thread_udp, server_thread_udp);
      if (config.servers[server_index].disable_tcp)
        {
          close (server->socket_tcp);
          server->socket_tcp = -1;
        }
      else
        start_server_thread (aux, server_index,
                             &server->thread_tcp, server_thread_tcp);
    }
  if (config.single_thread_udp)
    start_server_thread_udp_single (aux);

  /* Initialize libresolv.  */
  res_init ();

  int timeout = 1;
  {
    const char *timeout_str = getenv ("TEST_RES_TIMEOUT");
    if (timeout_str != NULL)
      {
        timeout = atoi (timeout_str);
        if (timeout <= 0)
          {
            printf ("error: invalid TEST_RES_TIMEOUT value: %s\n",
                    timeout_str);
            abort ();
          }
      }
  }

  /* Redirect queries to the server socket.  */
  if (verbose)
    {
      printf ("info: old timeout value: %d\n", _res.retrans);
      printf ("info: old retry attempt value: %d\n", _res.retry);
      printf ("info: old _res.options: 0x%lx\n", _res.options);
      printf ("info: old _res.nscount value: %d\n", _res.nscount);
      printf ("info: old _res.ndots value: %d\n", _res.ndots);
    }
  _res.retrans = timeout;
  _res.retry = 4;
  _res.nscount = config.nscount;
  _res.options = RES_INIT | RES_RECURSE | RES_DEFNAMES
    | RES_DNSRCH | RES_NOIP6DOTINT;
  _res.ndots = 1;
  if (verbose)
    {
      printf ("info: new timeout value: %d\n", _res.retrans);
      printf ("info: new retry attempt value: %d\n", _res.retry);
      printf ("info: new _res.options: 0x%lx\n", _res.options);
      printf ("info: new _res.nscount value: %d\n", _res.nscount);
      printf ("info: new _res.ndots value: %d\n", _res.ndots);
    }
  for (int server_index = 0; server_index < config.nscount; ++server_index)
    {
      _res.nsaddr_list[server_index] = aux->servers[server_index].address;
      if (verbose)
        {
          char buf[256];
          if (inet_ntop (AF_INET, &aux->servers[server_index].address.sin_addr,
                         buf, sizeof (buf)) == NULL)
            {
              printf ("error: inet_ntop: %m\n");
              abort ();
            }
          printf ("info: server %d: %s/%u\n",
                  server_index, buf,
                  htons (aux->servers[server_index].address.sin_port));
        }
    }

  set_search_path (config);

  return aux;
}

void
resolv_redirect_finish (struct resolv_aux *aux)
{
  res_close ();

  resolv_xlock (&aux->lock);
  if (aux->errors != 0)
    {
      printf ("error: aborting due to previous errors\n");
      abort ();
    }
  aux->termination_requested = true;
  resolv_xunlock (&aux->lock);

  /* Send triggers to exit the server threads.  */
  for (int server_index = 0; server_index < aux->config.nscount;
       ++server_index)
    {
      if (!aux->config.servers[server_index].disable_udp)
        {
          int sock = xsocket (SOCK_DGRAM);
          xsendto (sock, "", 1, 0,
                   (struct sockaddr *) &aux->servers[server_index].address,
                   sizeof (aux->servers[server_index].address));
          close (sock);
        }
      if (!aux->config.servers[server_index].disable_tcp)
        {
          int sock = xsocket (SOCK_STREAM);
          if (connect (sock,
                       (struct sockaddr *) &aux->servers[server_index].address,
                       sizeof (aux->servers[server_index].address)))
            {
              printf ("error: TCP connect: %m\n");
              abort ();
            }
          close (sock);
        }
    }

  if (aux->config.single_thread_udp)
    xjoin (aux->thread_udp_single);

  /* Wait for the server threads to terminate.  */
  for (int server_index = 0; server_index < aux->config.nscount;
       ++server_index)
    {
      if (!aux->config.servers[server_index].disable_udp)
        {
          if (!aux->config.single_thread_udp)
            xjoin (aux->servers[server_index].thread_udp);
          close (aux->servers[server_index].socket_udp);
        }
      if (!aux->config.servers[server_index].disable_tcp)
        {
          xjoin (aux->servers[server_index].thread_tcp);
          close (aux->servers[server_index].socket_tcp);
        }
    }

  resolv_xlock (&aux->lock);
  if (aux->errors != 0)
    {
      printf ("error: aborting due to previous errors\n");
      abort ();
    }
  resolv_xunlock (&aux->lock);

  free (aux);
}

/* Response builder. */

static void
response_push_pointer_to_free (struct response_builder *b, void *ptr)
{
  if (ptr == NULL)
    return;
  struct to_be_freed *e = xmalloc (sizeof (*e));
  *e = (struct to_be_freed) {b->to_be_freed, ptr};
  b->to_be_freed = e;
}

void
response_init (struct response_builder *b, struct response_flags flags)
{
  if (b->offset > 0)
    {
      printf ("error: response_init: called at offser %zu\n", b->offset);
      abort ();
    }
  if (b->query_length < 12)
    {
      printf ("error: response_init called for a query of size %zu\n",
              b->query_length);
      abort ();
    }
  if (flags.rcode > 15)
    {
      printf ("error: response_init: invalid RCODE %u\n", flags.rcode);
      abort ();
    }

  /* Copy the transaction ID.  */
  b->buffer[0] = b->query_buffer[0];
  b->buffer[1] = b->query_buffer[1];

  /* Initialize the flags.  */
  b->buffer[2] = 0x80;                       /* Mark as response.   */
  b->buffer[2] |= b->query_buffer[2] & 0x01; /* Copy the RD bit.  */
  if (flags.tc)
    b->buffer[2] |= 0x02;
  b->buffer[3] = 0x80 | flags.rcode; /* Always set RA.  */

  /* Fill in the initial section count values.  */
  b->buffer[4] = flags.qdcount >> 8;
  b->buffer[5] = flags.qdcount;
  b->buffer[6] = flags.ancount >> 8;
  b->buffer[7] = flags.ancount;
  b->buffer[8] = flags.nscount >> 8;
  b->buffer[9] = flags.nscount;
  b->buffer[10] = flags.adcount >> 8;
  b->buffer[11] = flags.adcount;

  b->offset = 12;
}

void
response_section (struct response_builder *b, enum response_section section)
{
  if (b->offset == 0)
    {
      printf ("error: response_section: must be called after response_init\n");
      abort ();
    }
  if (section < b->section)
    {
      printf ("error: response_section: cannot go back in section");
      abort ();
    }
  b->section = section;
}

static inline void
response_add_byte (struct response_builder *b, unsigned char ch)
{
  if (b->offset == max_response_length)
    {
      printf ("error: DNS response exceeds 64 KiB limit\n");
      abort ();
    }
  b->buffer[b->offset] = ch;
  ++b->offset;
}

static void
response_add_16 (struct response_builder *b, uint16_t val)
{
  response_add_byte (b, val >> 8);
  response_add_byte (b, val);
}

static void
response_count_increment (struct response_builder *b)
{
  unsigned offset = b->section;
  offset = 4 + 2 * offset;
  ++b->buffer[offset + 1];
  if (b->buffer[offset + 1] == 0)
    {
      /* Carry.  */
      ++b->buffer[offset];
      if (b->buffer[offset] == 0)
        {
          /* Overflow.  */
          printf ("error: too many records in section\n");
          abort ();
        }
    }
}

void
response_add_question (struct response_builder *b,
                       const char *name, uint16_t class, uint16_t type)
{
  if (b->offset == 0)
    {
      printf ("error: response_add_question: "
              "must be called after response_init\n");
      abort ();
    }
  if (b->section != section_question)
    {
      printf ("error: response_add_question: "
              "must be called in the question section\n");
      abort ();
    }

  response_add_name (b, name);
  response_add_16 (b, type);
  response_add_16 (b, class);

  response_count_increment (b);
}

void
response_add_name (struct response_builder *b,
                   const char *const origname)
{
  /* Normalized name.  */
  char *name;
  /* Normalized name with case preserved.  */
  char *name_case;
  {
    size_t namelen = strlen (origname);
    /* Remove trailing dots.  FIXME: Handle trailing quoted dots.  */
    while (namelen > 0 && origname[namelen - 1] == '.')
      --namelen;
    name = xmalloc (namelen + 1);
    name_case = xmalloc (namelen + 1);
    /* Copy and convert to lowercase.  FIXME: This needs to normalize
       escaping as well.  */
    for (size_t i = 0; i < namelen; ++i)
      {
        char ch = origname[i];
        name_case[i] = ch;
        if ('A' <= ch && ch <= 'Z')
          ch = ch - 'A' + 'a';
        name[i] = ch;
      }
    name[namelen] = 0;
    name_case[namelen] = 0;
  }
  char *name_start = name;
  char *name_case_start = name_case;

  bool compression = false;
  while (*name)
    {
      /* Search for a previous name we can reference.  */
      ENTRY new_entry =
        {
          .key = name,
          .data = (void *) (uintptr_t) b->offset,
        };

      /* If the label can be a compression target because it is at a
         reachable offset, add it to the hash table.  */
      ACTION action;
      if (b->offset < (1 << 12))
        action = ENTER;
      else
        action = FIND;

      /* Search for known compression offsets in the hash table.  */
      ENTRY *e;
      if (hsearch_r (new_entry, action, &e, &b->compression_offsets) == 0)
        {
          if (action == FIND && errno == ESRCH)
            /* Fall through.  */
            e = NULL;
          else
            {
              printf ("error: hsearch_r failure in name compression: %m\n");
              abort ();
            }
        }

      /* The name is known.  Reference the previous location.  */
      if (e != NULL && e->data != new_entry.data)
        {
          size_t old_offset = (uintptr_t) e->data;
          response_add_byte (b, 0xC0 | (old_offset >> 8));
          response_add_byte (b, old_offset);
          compression = true;
          break;
        }

      /* The name does not exist yet.  Write one label.  First, add
         room for the label length.  */
      size_t buffer_label_offset = b->offset;
      response_add_byte (b, 0);

      /* Copy the label.  */
      while (true)
        {
          char ch = *name_case;
          if (ch == '\0')
            break;
          ++name;
          ++name_case;
          if (ch == '.')
            break;
          /* FIXME: Handle escaping.  */
          response_add_byte (b, ch);
        }

      /* Patch in the label length.  */
      size_t label_length = b->offset - buffer_label_offset - 1;
      if (label_length == 0)
        {
          printf ("error: empty label in name compression: %s\n", origname);
          abort ();
        }
      if (label_length > 63)
        {
          printf ("error: label too long in name compression: %s\n", origname);
          abort ();
        }
      b->buffer[buffer_label_offset] = label_length;

      /* Continue with the tail of the name and the next label.  */
    }

  if (compression)
    {
      /* If we found an immediate match for the name, we have not put
         it into the hash table, and can free it immediately.  */
      if (name == name_start)
        free (name_start);
      else
        response_push_pointer_to_free (b, name_start);
    }
  else
    {
      /* Terminate the sequence of labels.  With compression, this is
         implicit in the compression reference.  */
      response_add_byte (b, 0);
      response_push_pointer_to_free (b, name_start);
    }

  free (name_case_start);
}

void
response_open_record (struct response_builder *b, const char *name,
                      uint16_t class, uint16_t type, uint32_t ttl)
{
  if (b->section == section_question)
    {
      printf ("error: response_open_record called in question section\n");
      abort ();
    }
  if (b->current_rdata_offset != 0)
    {
      printf ("error: response_open_record called with open record\n");
      abort ();
    }

  response_add_name (b, name);
  response_add_16 (b, type);
  response_add_16 (b, class);
  response_add_16 (b, ttl >> 16);
  response_add_16 (b, ttl);

  b->current_rdata_offset = b->offset;
  /* Add room for the RDATA length.  */
  response_add_16 (b, 0);
}


void
response_close_record (struct response_builder *b)
{
  size_t rdata_offset = b->current_rdata_offset;
  if (rdata_offset == 0)
    {
      printf ("error: response_close_record called without open record\n");
      abort ();
    }
  size_t rdata_length = b->offset - rdata_offset - 2;
  if (rdata_length > 65535)
    {
      printf ("error: RDATA length %zu exceeds limit\n", rdata_length);
      abort ();
    }
  b->buffer[rdata_offset] = rdata_length >> 8;
  b->buffer[rdata_offset + 1] = rdata_length;
  response_count_increment (b);
  b->current_rdata_offset = 0;
}

void
response_add_data (struct response_builder *b,
                   const void *data, size_t length)
{
  size_t remaining = max_response_length - b->offset;
  if (remaining < length)
    {
      printf ("error: response_add_data: not enough room for %zu bytes\n",
              length);
      abort ();
    }
  memcpy (b->buffer + b->offset, data, length);
  b->offset += length;
}

void
response_drop (struct response_builder *b)
{
  b->drop = true;
}

void
response_close (struct response_builder *b)
{
  b->close = true;
}

size_t
response_length (const struct response_builder *b)
{
  return b->offset;
}

unsigned char *
response_buffer (const struct response_builder *b)
{
  unsigned char *result = malloc (b->offset);
  if (result == NULL)
    {
      printf ("error: malloc: %m\n");
      abort ();
    }
  memcpy (result, b->buffer, b->offset);
  return result;
}

static struct response_builder *
response_builder_allocate
  (const unsigned char *query_buffer, size_t query_length)
{
  struct response_builder *b = xmalloc (sizeof (*b));
  *b = (struct response_builder)
    {
      .query_buffer = query_buffer,
      .query_length = query_length
    };
  if (hcreate_r (10000, &b->compression_offsets) == 0)
    {
      printf ("error: hcreate_r: %m\n");
      abort ();
    }
  return b;
}

static void
response_builder_free (struct response_builder *b)
{
  struct to_be_freed *current = b->to_be_freed;
  while (current != NULL)
    {
      struct to_be_freed *next = current->next;
      free (current->ptr);
      free (current);
      current = next;
    }
  hdestroy_r (&b->compression_offsets);
  free (b);
}

static const char *
format_address_family (int family)
{
  switch (family)
    {
    case AF_INET:
      return "INET";
    case AF_INET6:
      return "INET6";
    }
  printf ("error: invalid address family: %d\n", family);
  abort ();
}

static int
address_length (int family)
{
  switch (family)
    {
    case AF_INET:
      return 4;
    case AF_INET6:
      return 16;
    }
  printf ("error: invalid address family: %d\n", family);
  abort ();
}

static size_t
socket_address_length (int family)
{
  switch (family)
    {
    case AF_INET:
      return sizeof (struct sockaddr_in);
    case AF_INET6:
      return sizeof (struct sockaddr_in6);
    }
  printf ("error: invalid address family: %d\n", family);
  abort ();
}

char *
format_hostent (struct hostent *h)
{
  int herrno_copy = h_errno;

  struct resolv_memstream mem;
  resolv_xopen_memstream (&mem);
  if (h == NULL)
    {
      const char *errstr;
      switch (herrno_copy)
        {
        case HOST_NOT_FOUND:
          errstr = "HOST_NOT_FOUND";
          break;
        case NO_ADDRESS:
          errstr = "NO_ADDRESS";
          break;
        case NO_RECOVERY:
          errstr = "NO_RECOVERY";
          break;
        case TRY_AGAIN:
          errstr = "TRY_AGAIN";
          break;
        default:
          printf ("error: invalid h_errno value: %d\n", herrno_copy);
          abort ();
        }
      fprintf (mem.out, "error: %s\n", errstr);
    }
  else
    {
      if (h->h_length !=  address_length (h->h_addrtype))
        {
          printf ("error: invalid address length %d for %s\n",
                  h->h_length, format_address_family (h->h_addrtype));
          abort ();
        }
      fprintf (mem.out, "name: %s\n", h->h_name);
      for (char **alias = h->h_aliases; *alias != NULL; ++alias)
        fprintf (mem.out, "alias: %s\n", *alias);
      for (char **addr = h->h_addr_list; *addr != NULL; ++addr)
        {
          char buf[128];
          if (inet_ntop (h->h_addrtype, *addr, buf, sizeof (buf)) == NULL)
            {
              printf ("error: inet_ntop failed: %m\n");
              abort ();
            }
          fprintf (mem.out, "address: %s\n", buf);
        }
    }
  resolv_xfclose_memstream (&mem);
  return mem.buffer;
}

static char *
write_to_temp_file (const char *prefix, const char *str)
{
  char *name;
  if (asprintf (&name, "/tmp/tst-resolv-%s-XXXXXX", prefix) < 0)
    {
      printf ("error: asprintf: %m\n");
      abort ();
    }
  int fd = mkstemp (name);
  if (fd < 0)
    {
      printf ("error: mkstemp: %m\n");
      abort ();
    }
  struct iovec buffers = {(char *) str, strlen (str)};
  xwritev_fully (fd, &buffers, 1);
  if (close (fd) != 0)
    {
      printf ("error: close: %m\n");
      abort ();
    }
  return name;
}

static void
run_diff (const char *left_label, const char *left,
          const char *right_label, const char *right)
{
  if (fflush (stdout) != 0)
    {
      printf ("error: fflush: %m\n");
      abort ();
    }

  char *left_path = write_to_temp_file ("left-diff", left);
  char *right_path = write_to_temp_file ("right-diff", right);

  pid_t pid = fork ();
  if (pid < 0)
    {
      printf ("error: fork: %m\n");
      unlink (right_path);
      unlink (left_path);
      abort ();
    }
  else if (pid == 0)
    {
      execlp ("diff", "diff", "-u",
              "--label", left_label, "--label", right_label,
              "--", left_path, right_path,
              NULL);
      _exit (17);
    }
  else
    {
      int status;
      int ret = waitpid (pid, &status, 0);
      int error = errno;
      unlink (right_path);
      unlink (left_path);
      errno = error;
      if (ret < 0)
        {
          printf ("error: waitpid: %m\n");
          abort ();
        }
      if (!WIFEXITED (status) || WEXITSTATUS (status) != 1)
        {
          printf ("warning: could not run diff, exit status: %d\n"
                  "*** %s ***\n%s\n"
                  "*** %s ***\n%s\n",
                  status, left_label, left, right_label, right);
        }
    }

  free (right_path);
  free (left_path);
}

void
check_hostent (struct hostent *h, const char *expected)
{
  char *formatted = format_hostent (h);
  if (strcmp (formatted, expected) != 0)
    {
      resolv_xlock (&errors_lock);
      ++errors;
      resolv_xunlock (&errors_lock);
      printf ("error: hostent comparison failure\n");
      run_diff ("expected", expected,
                "actual", formatted);
    }
  free (formatted);
}

static void
format_ai_flags (FILE *out, struct addrinfo *ai, int flag, const char *name,
                 int * flags_printed)
{
  if ((ai->ai_flags & flag) != 0)
    fprintf (out, " %s", name);
  *flags_printed |= flag;
}

static void
format_ai_one (FILE *out, struct addrinfo *ai, int *flags)
{
  /* ai_flags */
  if (ai->ai_flags != *flags)
    {
      fprintf (out, "flags:");
      int flags_printed = 0;
#define FLAG(flag) format_ai_flags (out, ai, flag, #flag, &flags_printed)
      FLAG (AI_PASSIVE);
      FLAG (AI_CANONNAME);
      FLAG (AI_NUMERICHOST);
      FLAG (AI_V4MAPPED);
      FLAG (AI_ALL);
      FLAG (AI_ADDRCONFIG);
      FLAG (AI_IDN);
      FLAG (AI_CANONIDN);
      FLAG (AI_IDN_ALLOW_UNASSIGNED);
      FLAG (AI_IDN_USE_STD3_ASCII_RULES);
      FLAG (AI_NUMERICSERV);
#undef FLAG
      int remaining = ai->ai_flags & ~flags_printed;
      if (remaining != 0)
        fprintf (out, " %08x", remaining);
      fprintf (out, "\n");
      *flags = ai->ai_flags;
    }

  {
    char type_buf[32];
    const char *type_str;
    char proto_buf[32];
    const char *proto_str;

    /* ai_socktype */
    switch (ai->ai_socktype)
      {
      case SOCK_RAW:
        type_str = "RAW";
        break;
      case SOCK_DGRAM:
        type_str = "DGRAM";
        break;
      case SOCK_STREAM:
        type_str = "STREAM";
        break;
      default:
        snprintf (type_buf, sizeof (type_buf), "%d", ai->ai_socktype);
        type_str = type_buf;
      }

    /* ai_protocol */
    switch (ai->ai_protocol)
      {
      case IPPROTO_IP:
        proto_str = "IP";
        break;
      case IPPROTO_UDP:
        proto_str = "UDP";
        break;
      case IPPROTO_TCP:
        proto_str = "TCP";
        break;
      default:
        snprintf (proto_buf, sizeof (proto_buf), "%d", ai->ai_protocol);
        proto_str = proto_buf;
      }
    fprintf (out, "address: %s/%s", type_str, proto_str);
  }

  /* ai_addrlen */
  if (ai->ai_addrlen != socket_address_length (ai->ai_family))
    {
      printf ("error: invalid address length %d for %s\n",
              ai->ai_addrlen, format_address_family (ai->ai_family));
      abort ();
    }

  /* ai_addr */
  {
    char buf[128];
    uint16_t port;
    const char *ret;
    switch (ai->ai_family)
      {
      case AF_INET:
        {
          struct sockaddr_in *sin = (struct sockaddr_in *) ai->ai_addr;
          ret = inet_ntop (AF_INET, &sin->sin_addr, buf, sizeof (buf));
          port = sin->sin_port;
        }
        break;
      case AF_INET6:
        {
          struct sockaddr_in6 *sin = (struct sockaddr_in6 *) ai->ai_addr;
          ret = inet_ntop (AF_INET6, &sin->sin6_addr, buf, sizeof (buf));
          port = sin->sin6_port;
        }
        break;
      default:
        printf ("error: invalid address family: %d\n", ai->ai_family);
        abort ();
      }
    if (ret == NULL)
      {
        printf ("error: ntop failed: %m\n");
        abort ();
      }
    fprintf (out, " %s %u\n", buf, ntohs (port));
  }

  /* ai_canonname */
  if (ai->ai_canonname != NULL)
    fprintf (out, "canonname: %s\n", ai->ai_canonname);
}

/* Format all the addresses in one address family.  */
static void
format_ai_family (FILE *out, struct addrinfo *ai, int family, int *flags)
{
  while (ai)
    {
      if (ai->ai_family == family)
        format_ai_one (out, ai, flags);
      ai = ai->ai_next;
    }
}

char *
format_addrinfo (struct addrinfo *ai, int ret)
{
  int errno_copy = errno;

  struct resolv_memstream mem;
  resolv_xopen_memstream (&mem);
  if (ret != 0)
    {
      fprintf (mem.out, "error: %s\n", gai_strerror (ret));
      if (ret == EAI_SYSTEM)
        {
          errno = errno_copy;
          fprintf (mem.out, "error: %m\n");
        }
    }
  else
    {
      int flags = 0;
      format_ai_family (mem.out, ai, AF_INET, &flags);
      format_ai_family (mem.out, ai, AF_INET6, &flags);
    }

  resolv_xfclose_memstream (&mem);
  return mem.buffer;
}

void
check_addrinfo (struct addrinfo *ai, int ret, const char *expected)
{
  char *formatted = format_addrinfo (ai, ret);
  if (strcmp (formatted, expected) != 0)
    {
      resolv_xlock (&errors_lock);
      ++errors;
      resolv_xunlock (&errors_lock);
      printf ("error: addrinfo comparison failure\n");
      run_diff ("expected", expected, "actual", formatted);
    }
  free (formatted);
}

struct in_buffer
{
  const unsigned char *data;
  size_t size;
};

static inline bool
extract_8 (struct in_buffer *in, unsigned char *value)
{
  if (in->size == 0)
    return false;
  *value = in->data[0];
  ++in->data;
  --in->size;
  return true;
}

static inline bool
extract_16 (struct in_buffer *in, unsigned short *value)
{
  if (in->size < 2)
    return false;
  *value = (in->data[0] << 8) | in->data[1];
  in->data += 2;
  in->size -= 2;
  return true;
}

static inline bool
extract_32 (struct in_buffer *in, unsigned *value)
{
  if (in->size < 4)
    return false;
  unsigned a = in->data[0];
  unsigned b = in->data[1];
  unsigned c = in->data[2];
  unsigned d = in->data[3];
  *value = (a << 24) | (b << 16) | (c << 8) | d;
  in->data += 4;
  in->size -= 4;
  return true;
}

static inline bool
extract_bytes (struct in_buffer *in, size_t length, struct in_buffer *value)
{
  if (in->size < length)
    return false;
  *value = (struct in_buffer) {in->data, length};
  in->data += length;
  in->size -= length;
  return true;
}

struct dname
{
  char name[MAXDNAME + 1];
};

static bool
extract_name (struct in_buffer full, struct in_buffer *in, struct dname *value)
{
  const unsigned char *full_end = full.data + full.size;
  /* Sanity checks; these indicate buffer misuse.  */
  if (in->data < full.data || in->data > full_end
      || in->size > (size_t) (full_end - in->data))
    abort ();
  int ret = dn_expand (full.data, full_end, in->data,
                       value->name, sizeof (value->name));
  if (ret < 0)
    return false;
  in->data += ret;
  in->size -= ret;
  return true;
}

char *
format_packet (const unsigned char *buffer, size_t length)
{
  struct in_buffer full = {buffer, length};
  struct in_buffer in = full;
  struct resolv_memstream mem;
  resolv_xopen_memstream (&mem);

  unsigned short txnid;
  unsigned short flags;
  unsigned short qdcount;
  unsigned short ancount;
  unsigned short nscount;
  unsigned short adcount;
  if (!(extract_16 (&in, &txnid)
        && extract_16 (&in, &flags)
        && extract_16 (&in, &qdcount)
        && extract_16 (&in, &ancount)
        && extract_16 (&in, &nscount)
        && extract_16 (&in, &adcount)))
    {
      fprintf (mem.out, "error: could not parse DNS header\n");
      goto out;
    }
  if (qdcount != 1)
    {
      fprintf (mem.out, "error: question count is %d, not 1\n", qdcount);
      goto out;
    }
  struct dname qname;
  if (!extract_name (full, &in, &qname))
    {
      fprintf (mem.out, "error: malformed QNAME\n");
      goto out;
    }
  unsigned short qtype;
  unsigned short qclass;
  if (!(extract_16 (&in, &qtype)
        && extract_16 (&in, &qclass)))
    {
      fprintf (mem.out, "error: malformed question\n");
      goto out;
    }
  if (qtype != T_A && qtype != T_AAAA && qtype != T_PTR)
    {
      fprintf (mem.out, "error: unsupported QTYPE %d\n", qtype);
      goto out;
    }

  fprintf (mem.out, "name: %s\n", qname.name);

  for (int i = 0; i < ancount; ++i)
    {
      struct dname rname;
      if (!extract_name (full, &in, &rname))
        {
          fprintf (mem.out, "error: malformed record name\n");
          goto out;
        }
      unsigned short rtype;
      unsigned short rclass;
      unsigned ttl;
      unsigned short rdlen;
      struct in_buffer rdata;
      if (!(extract_16 (&in, &rtype)
            && extract_16 (&in, &rclass)
            && extract_32 (&in, &ttl)
            && extract_16 (&in, &rdlen)
            && extract_bytes (&in, rdlen, &rdata)))
        {
          fprintf (mem.out, "error: malformed record header\n");
          goto out;
        }
      /* Skip non-matching record types.  */
      if (rtype != qtype || rclass != qclass)
        continue;
      switch (rtype)
        {
        case T_A:
          if (rdlen == 4)
              fprintf (mem.out, "address: %d.%d.%d.%d\n",
                       rdata.data[0],
                       rdata.data[1],
                       rdata.data[2],
                       rdata.data[3]);
          else
            fprintf (mem.out, "error: A record of size %d: %s\n", rdlen, rname.name);
          break;
        case T_AAAA:
          {
            char buf[100];
            if (inet_ntop (AF_INET6, rdata.data, buf, sizeof (buf)) == NULL)
              fprintf (mem.out, "error: AAAA record decoding failed: %m\n");
            else
              fprintf (mem.out, "address: %s\n", buf);
          }
          break;
        case T_CNAME:
        case T_PTR:
          {
            struct dname name;
            if (extract_name (full, &in, &name))
              fprintf (mem.out, "name: %s\n", name.name);
            else
              fprintf (mem.out, "error: malformed PTR record\n");
          }
        }
    }

 out:
  resolv_xfclose_memstream (&mem);
  return mem.buffer;
}

void
check_packet (const unsigned char *buffer, size_t length,
              const char *expected)
{
  char *formatted = format_packet (buffer, length);
  if (strcmp (formatted, expected) != 0)
    {
      resolv_xlock (&errors_lock);
      ++errors;
      resolv_xunlock (&errors_lock);
      printf ("error: packet comparison failure\n");
      run_diff ("expected", expected, "actual", formatted);
    }
  free (formatted);
}

/* open_memstream-related wrappers.  */

void
resolv_xopen_memstream (struct resolv_memstream *stream)
{
  *stream = (struct resolv_memstream) {};
  stream->out = open_memstream (&stream->buffer, &stream->length);
}

void
resolv_xfclose_memstream (struct resolv_memstream *stream)
{
  if (ferror (stream->out))
    {
      printf ("error: memstream write: %m\n");
      abort ();
    }
  if (fclose (stream->out) != 0)
    {
      printf ("error: memstream close: %m\n");
      abort ();
    }
  stream->out = NULL;
}

static void
__attribute__ ((destructor))
report_error_count (void)
{
  int errors_local;
  resolv_xlock (&errors_lock);
  errors_local = errors;
  resolv_xunlock (&errors_lock);
  if (errors_local > 0)
    {
      printf ("error: %d errors detected\n", errors);
      abort ();
    }
}
