/* Test DNS reverse lookups via gethostbyaddr and getnameinfo.
   Copyright (C) 2016 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.  */

#include "tst-resolv-aux.h"

#include <resolv.h>
#include <netdb.h>
#include <stdlib.h>
#include <string.h>

enum answer
  {
    answer_none,
    answer_nxdomain,
    answer_single,
    answer_double,
    answer_multi_small,
    answer_multi_large,
    answer_underscore,
    answer_meta,
    answer_meta_chain,
    answer_element_count        /* number of enum elements above */
  };


struct dname
{
  /* Failures start with 255 bytes for some reason. */
  char name[254];
};

static void
format_name (int n, struct dname *name)
{
  memset (name->name, 'a', sizeof (name->name));
  name->name[sizeof (name->name) - 1] = '\0';
  name->name[3] = '.';
  name->name[64] = '.';
  name->name[128] = '.';
  name->name[192] = '.';

  /* Fill in discriminating prefix.  */
  name->name[2] = 'a' + (n % 26);
  n /= 26;
  name->name[1] = 'a' + (n % 26);
  n /= 26;
  name->name[0] = 'a' + (n % 26);
}

static void
add_expected_names (FILE *out, int count)
{
  /* nss_dns only extracts a single host name from the answer.  */
  struct dname name;
  format_name (count, &name);
  fprintf (out, "name: %s\n", name.name);
}

static int
record_count_for_answer (int family, enum answer answer)
{
  switch (answer)
    {
    case answer_none:
    case answer_nxdomain:
    case answer_underscore:
    case answer_meta:
    case answer_meta_chain:
    case answer_element_count:
      return -1;
    case answer_single:
      return 1;
    case answer_double:
      return 2;
    case answer_multi_small:
      switch (family)
        {
        case AF_INET:
          return 12;
        case AF_INET6:
          return 9;
        }
      return -1;
    case answer_multi_large:
      switch (family)
        {
        case AF_INET:
          return 3624;
        case AF_INET6:
          return 3622;
        }
      return -1;
    }
  abort ();
}

static char *
format_expected (int family, enum answer answer)
{
  struct resolv_memstream stream;
  resolv_xopen_memstream (&stream);

  switch (answer)
    {
    case answer_none:
      family = -1;
      fprintf (stream.out, "error: NO_ADDRESS\n");
      break;
    case answer_nxdomain:
      family = -1;
      fprintf (stream.out, "error: HOST_NOT_FOUND\n");
      break;
    case answer_single:
    case answer_double:
    case answer_multi_small:
    case answer_multi_large:
      add_expected_names (stream.out,
                          record_count_for_answer (family, answer));
      break;
    case answer_underscore:
      fprintf (stream.out, "name: host_name.example\n");
      break;
    case answer_meta:
      family = -1;
      fprintf (stream.out, "error: NO_RECOVERY\n");
      break;
    case answer_meta_chain:
      fprintf (stream.out, "name: hostname.example\n");
      break;
    case answer_element_count:
      break;
    }

  switch (family)
    {
    case AF_INET:
      fprintf (stream.out, "address: 192.0.2.%d\n", answer);
      break;
    case AF_INET6:
      fprintf (stream.out, "address: 2001:db8::%x\n", answer);
      break;
    }

  resolv_xfclose_memstream (&stream);
  return stream.buffer;
}

static char *
format_expected_gni (int family, enum answer answer)
{
  switch (answer)
    {
    case answer_none:
    case answer_nxdomain:
    case answer_meta:
      {
        char *name;
        int ret;
        switch (family)
          {
          case AF_INET:
            ret = asprintf (&name, "192.0.2.%d", answer);
            break;
          case AF_INET6:
            if (answer == 0)
              ret = asprintf (&name, "2001:db8::");
            else
              ret = asprintf (&name, "2001:db8::%x", answer);
            break;
          default:
            abort ();
          }
        if (ret < 0)
          {
            printf ("error: asprintf: %m\n");
            abort ();
          }
        return name;
      }
      break;
    case answer_single:
    case answer_double:
    case answer_multi_small:
    case answer_multi_large:
      {
        struct dname name;
        format_name (record_count_for_answer (family, answer), &name);
        char *result = strdup (name.name);
        if (result == NULL)
          {
            printf ("error: strdup: %m\n");
            abort ();
          }
        return result;
      }
      break;
    case answer_underscore:
      {
        char * result = strdup ("host_name.example");
        if (result == NULL)
          {
            printf ("error: strdup: %m\n");
            abort ();
          }
        return result;
      }
    case answer_meta_chain:
      {
        char * result = strdup ("hostname.example");
        if (result == NULL)
          {
            printf ("error: strdup: %m\n");
            abort ();
          }
        return result;
      }
    case answer_element_count:
      break;
    }
  abort ();
}

static void
add_ptr_records (struct response_builder *b, const char *qname, int count)
{
  for (int i = 0; i < count; ++i)
    {
      struct dname name;
      format_name (count - i, &name);
      response_open_record (b, qname, C_IN, T_PTR, 0);
      response_add_name (b, name.name);
      response_close_record (b);
    }
}

static void
response (const struct response_context *ctx,
          struct response_builder *b,
          const char *qname, uint16_t qclass, uint16_t qtype)
{
  if (qtype != T_PTR)
    {
      printf ("error: invalid QTYPE %d\n", qtype);
      abort ();
    }
  enum answer answer;
  {
    int a;
    if (sscanf (qname, "%d", &a) != 1)
      {
        printf ("error: cannot parse query: %s\n", qname);
        abort ();
      }
    answer = a;
  }

  /* Perform TCP fallback for large answers.  */
  if (answer == answer_multi_large && !ctx->tcp)
    {
      struct response_flags flags = {.tc = true};
      response_init (b, flags);
      response_add_question (b, qname, qclass, qtype);
      return;
    }

  /* Special case for NXDOMAIN response.  */
  if (answer == answer_nxdomain)
    {
      struct response_flags flags = {.rcode = ns_r_nxdomain};
      response_init (b, flags);
      response_add_question (b, qname, qclass, qtype);
      return;
    }

  struct response_flags flags = {};
  response_init (b, flags);
  response_add_question (b, qname, qclass, qtype);
  response_section (b, section_answer);

  /* Determine the address family.  */
  int family;
  if (strstr (qname, "in-addr.arpa") == NULL)
    family = AF_INET6;
  else
    family = AF_INET;

  switch (answer)
    {
    case answer_none:
      break;
    case answer_single:
    case answer_double:
    case answer_multi_small:
    case answer_multi_large:
      add_ptr_records (b, qname, record_count_for_answer (family, answer));
      break;
    case answer_underscore:
      response_open_record (b, qname, C_IN, T_PTR, 0);
      response_add_name (b, "host_name.example");
      response_close_record (b);
      break;
    case answer_meta:
      response_open_record (b, qname, C_IN, T_PTR, 0);
      response_add_name (b, "host`name.example");
      response_close_record (b);
      break;
    case answer_meta_chain:
      response_open_record (b, qname, C_IN, T_CNAME, 0);
      response_add_name (b, "cidr/25.example");
      response_close_record (b);
      response_open_record (b, "cidr/25.example", C_IN, T_PTR, 0);
      response_add_name (b, "hostname.example");
      response_close_record (b);
      break;
    case answer_nxdomain:
    case answer_element_count:
      abort ();
    }
}

int
main (void)
{
  struct redirect_config config =
    {
      .response_callback = response,
    };
  struct resolv_aux *aux = resolv_redirect (config);

  {
    char address[4] = {192, 0, 2, 0};
    for (int answer = 0; answer < answer_element_count; ++answer)
      {
        address[3] = answer;
        struct hostent *e = gethostbyaddr (address, sizeof (address), AF_INET);
        char *expected = format_expected (AF_INET, answer);
        check_hostent (e, expected);
        free (expected);

        expected = format_expected_gni (AF_INET, answer);
        size_t host_length = sizeof (struct dname);
        char *host = malloc (host_length);
        if (host == NULL)
          {
            printf ("error: malloc: %m\n");
            abort ();
          }
        struct sockaddr_in saddr =
          {
            .sin_family = AF_INET
          };
        memcpy (&saddr.sin_addr, address, sizeof (address));
        int ret = getnameinfo ((struct sockaddr *) &saddr, sizeof (saddr),
                               host, host_length, NULL, 0, 0);
        if (ret != 0)
          {
            printf ("error: getnameinfo: %s\n", gai_strerror (ret));
            abort ();
          }
        if (strcmp (host, expected) != 0)
          {
            printf ("error: getnameinfo difference for %d\n", answer);
            printf ("expected: %s\n", expected);
            printf ("actual: %s\n", host);
            abort ();
          }
        free (host);
        free (expected);
      }
  }

  {
    char address[16]
      = {0x20, 0x01, 0xd, 0xb8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    for (int answer = 0; answer < answer_element_count; ++answer)
      {
        address[15] = answer;
        struct hostent *e = gethostbyaddr (address, sizeof (address), AF_INET6);
        char *expected = format_expected (AF_INET6, answer);
        check_hostent (e, expected);
        free (expected);

        expected = format_expected_gni (AF_INET6, answer);
        size_t host_length = sizeof (struct dname);
        char *host = malloc (host_length);
        if (host == NULL)
          {
            printf ("error: malloc: %m\n");
            abort ();
          }
        struct sockaddr_in6 saddr =
          {
            .sin6_family = AF_INET6
          };
        memcpy (&saddr.sin6_addr, address, sizeof (address));
        int ret = getnameinfo ((struct sockaddr *) &saddr, sizeof (saddr),
                               host, host_length, NULL, 0, 0);
        if (ret != 0)
          {
            printf ("error: getnameinfo: %s\n", gai_strerror (ret));
            abort ();
          }
        if (strcmp (host, expected) != 0)
          {
            printf ("error: getnameinfo difference for %d\n", answer);
            printf ("expected: %s\n", expected);
            printf ("actual: %s\n", host);
            abort ();
          }
        free (host);
        free (expected);
      }
  }

  resolv_redirect_finish (aux);
}
