/* Test NSS retry behavior with differing response sizes.
   Copyright (C) 2016 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.  */


/* This test exercises response size handling and retry behavior on
   the getaddrinfo/nss_dns boundary.  It does not perform any response
   manipulation.  This test demonstrates an excessive number of
   queries with very large DNS responses containing many
   addresses.  */

#include "tst-resolv-aux.h"

#include <errno.h>
#include <netdb.h>
#include <pthread.h>
#include <resolv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;

static int address_count;
static int response_number;

static void
response (const struct response_context *ctx,
          struct response_builder *b,
          const char *qname, uint16_t qclass, uint16_t qtype)
{
  if (qname == NULL)
    {
      printf ("error: unparseable query\n");
      abort ();
    }

  struct response_flags flags = {};
  response_init (b, flags);
  response_add_question (b, qname, qclass, qtype);

  response_section (b, section_answer);

  /* The number of addresses (in the additional section) for the name
     server record (in the authoritative section).  */
  int address_count_local;
  resolv_xlock (&lock);
  address_count_local = address_count;
  ++response_number;
  resolv_xunlock (&lock);

  for (int i = 0; i < address_count_local; ++i)
    {
      response_open_record (b, qname, qclass, qtype, 0);
      switch (qtype)
        {
        case T_A:
          {
            char ipv4[4] =
              {
                response_number,
                i >> 8, i,
                ctx->tcp + 2 * ctx->server_index
              };
            response_add_data (b, &ipv4, sizeof (ipv4));
          }
          break;
        case T_AAAA:
          {
            char ipv6[16]
              = {0x20, 0x01, 0xd, 0xb8, 0, 0, 0, 0, 0, 0,
                 response_number >> 8, response_number,
                 i >> 8, i,
                 0, ctx->tcp + 2 * ctx->server_index};
            response_add_data (b, &ipv6, sizeof (ipv6));
          }
          break;
        default:
          printf ("error: unexpected QTYPE: %s/%u/%u\n",
                  qname, qclass, qtype);
          abort ();
        }
      response_close_record (b);
    }
}

/* Obtain the answer number used from the answer.  We do not know when
   getaddrinfo switches buffer sizes and needs a new buffer, which is
   why we have to get this information from the answer.  */
static int
response_number_from_answer (struct addrinfo *ai, int ret)
{
  if (ret != 0)
    return -1;
  for (; ai != NULL; ai = ai->ai_next)
    {
      /* Extract the response number from the IN/A response, which we
         assume comes first.  */
      if (ai->ai_family == AF_INET)
        {
          struct sockaddr_in *sin = (struct sockaddr_in *) ai->ai_addr;
          return ((uint32_t) ntohl (sin->sin_addr.s_addr)) >> 24;
        }
    }
  return -1;
}

static char *
expected_result (int number, int address_count)
{
  int port = 80;
  struct resolv_memstream mem;
  resolv_xopen_memstream (&mem);
  int server_index = 0;
  for (int i = 0; i < address_count; ++i)
    fprintf (mem.out, "address: STREAM/TCP %u.%u.%u.%u %u\n",
             number & 0xff,
             (i >> 8) & 0xff, i & 0xff,
             2 * server_index, port);
  for (int i = 0; i < address_count; ++i)
    fprintf (mem.out, "address: STREAM/TCP 2001:db8::%x:%x:%x %u\n",
             (number + 1) & 0xffff,
             i & 0xffff,
             2 * server_index, port);
  resolv_xfclose_memstream (&mem);
  return mem.buffer;
}

static void
test_different_sizes (void)
{
  struct addrinfo hints =
    {
      .ai_family = AF_UNSPEC,
      .ai_socktype = SOCK_STREAM,
      .ai_protocol = IPPROTO_TCP,
    };
  struct addrinfo *ai;
  char *expected;
  int ret;
  int previous_number = 0;

  for (int i = 1; i < 500; ++i)
    {
      resolv_xlock (&lock);
      response_number = 0;
      address_count = i;
      resolv_xunlock (&lock);

      char name[40];
      snprintf (name, sizeof (name), "%04d.example", i);

      ret = getaddrinfo (name, "80", &hints, &ai);
      int number = response_number_from_answer (ai, ret);
      expected = expected_result (number, i);
      check_addrinfo (ai, ret, expected);
      if (number > 0)
        {
          if (number < previous_number)
            {
              printf ("error: from address count %d to %d, the number of"
                      " responses\n", i - 1, i);
              printf ("error: decreased from %d to %d\n",
                      previous_number, number);
              abort ();
            }
          previous_number = number;
        }
      if (ret == 0)
        freeaddrinfo (ai);
      free (expected);
    }
}

int
main (void)
{
  struct resolv_aux *aux = resolv_redirect
    ((struct redirect_config)
     {
       .response_callback = response
     });

  test_different_sizes ();

  _res.options |= RES_SNGLKUP;
  test_different_sizes ();

  _res.options |= RES_SNGLKUPREOP;
  test_different_sizes ();

  resolv_redirect_finish (aux);
}
